# mypy: ignore-errors
"""Generated protobuf code for EntDB SDK.

Do not edit manually - regenerate with scripts/generate_proto.sh

This module is internal to the SDK. Users should not import from here.
"""

from .entdb_pb2 import (
    CreateEdgeOp,
    CreateNodeOp,
    DeleteEdgeOp,
    DeleteNodeOp,
    Edge,
    # Execute
    ExecuteAtomicRequest,
    ExecuteAtomicResponse,
    # Edges
    GetEdgesRequest,
    GetEdgesResponse,
    GetMailboxRequest,
    GetMailboxResponse,
    # Nodes
    GetNodeRequest,
    GetNodeResponse,
    GetNodesRequest,
    GetNodesResponse,
    # Receipt
    GetReceiptStatusRequest,
    GetReceiptStatusResponse,
    GetSchemaRequest,
    GetSchemaResponse,
    # Health/Schema
    HealthRequest,
    HealthResponse,
    # Tenants
    ListMailboxUsersRequest,
    ListMailboxUsersResponse,
    ListTenantsRequest,
    ListTenantsResponse,
    MailboxItem,
    MailboxSearchResult,
    Node,
    NodeRef,
    Operation,
    QueryNodesRequest,
    QueryNodesResponse,
    Receipt,
    ReceiptStatus,
    # Common
    RequestContext,
    # Mailbox
    SearchMailboxRequest,
    SearchMailboxResponse,
    TenantInfo,
    TypedNodeRef,
    UpdateNodeOp,
)
from .entdb_pb2_grpc import EntDBServiceStub

__all__ = [
    "RequestContext",
    "Receipt",
    "ExecuteAtomicRequest",
    "ExecuteAtomicResponse",
    "Operation",
    "CreateNodeOp",
    "UpdateNodeOp",
    "DeleteNodeOp",
    "CreateEdgeOp",
    "DeleteEdgeOp",
    "NodeRef",
    "TypedNodeRef",
    "GetReceiptStatusRequest",
    "GetReceiptStatusResponse",
    "ReceiptStatus",
    "GetNodeRequest",
    "GetNodeResponse",
    "GetNodesRequest",
    "GetNodesResponse",
    "QueryNodesRequest",
    "QueryNodesResponse",
    "Node",
    "GetEdgesRequest",
    "GetEdgesResponse",
    "Edge",
    "SearchMailboxRequest",
    "SearchMailboxResponse",
    "MailboxSearchResult",
    "GetMailboxRequest",
    "GetMailboxResponse",
    "MailboxItem",
    "HealthRequest",
    "HealthResponse",
    "GetSchemaRequest",
    "GetSchemaResponse",
    "ListTenantsRequest",
    "ListTenantsResponse",
    "TenantInfo",
    "ListMailboxUsersRequest",
    "ListMailboxUsersResponse",
    "EntDBServiceStub",
]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import entdb_pb2 as entdb__pb2


class EntDBServiceStub(object):
    """EntDB Service - Main API for database operations
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ExecuteAtomic = channel.unary_unary(
                '/entdb.v1.EntDBService/ExecuteAtomic',
                request_serializer=entdb__pb2.ExecuteAtomicRequest.SerializeToString,
                response_deserializer=entdb__pb2.ExecuteAtomicResponse.FromString,
                )
        self.GetReceiptStatus = channel.unary_unary(
                '/entdb.v1.EntDBService/GetReceiptStatus',
                request_serializer=entdb__pb2.GetReceiptStatusRequest.SerializeToString,
                response_deserializer=entdb__pb2.GetReceiptStatusResponse.FromString,
                )
        self.GetNode = channel.unary_unary(
                '/entdb.v1.EntDBService/GetNode',
                request_serializer=entdb__pb2.GetNodeRequest.SerializeToString,
                response_deserializer=entdb__pb2.GetNodeResponse.FromString,
                )
        self.GetNodes = channel.unary_unary(
                '/entdb.v1.EntDBService/GetNodes',
                request_serializer=entdb__pb2.GetNodesRequest.SerializeToString,
                response_deserializer=entdb__pb2.GetNodesResponse.FromString,
                )
        self.QueryNodes = channel.unary_unary(
                '/entdb.v1.EntDBService/QueryNodes',
                request_serializer=entdb__pb2.QueryNodesRequest.SerializeToString,
                response_deserializer=entdb__pb2.QueryNodesResponse.FromString,
                )
        self.GetEdgesFrom = channel.unary_unary(
                '/entdb.v1.EntDBService/GetEdgesFrom',
                request_serializer=entdb__pb2.GetEdgesRequest.SerializeToString,
                response_deserializer=entdb__pb2.GetEdgesResponse.FromString,
                )
        self.GetEdgesTo = channel.unary_unary(
                '/entdb.v1.EntDBService/GetEdgesTo',
                request_serializer=entdb__pb2.GetEdgesRequest.SerializeToString,
                response_deserializer=entdb__pb2.GetEdgesResponse.FromString,
                )
        self.SearchMailbox = channel.unary_unary(
                '/entdb.v1.EntDBService/SearchMailbox',
                request_serializer=entdb__pb2.SearchMailboxRequest.SerializeToString,
                response_deserializer=entdb__pb2.SearchMailboxResponse.FromString,
                )
        self.GetMailbox = channel.unary_unary(
                '/entdb.v1.EntDBService/GetMailbox',
                request_serializer=entdb__pb2.GetMailboxRequest.SerializeToString,
                response_deserializer=entdb__pb2.GetMailboxResponse.FromString,
                )
        self.Health = channel.unary_unary(
                '/entdb.v1.EntDBService/Health',
                request_serializer=entdb__pb2.HealthRequest.SerializeToString,
                response_deserializer=entdb__pb2.HealthResponse.FromString,
                )
        self.GetSchema = channel.unary_unary(
                '/entdb.v1.EntDBService/GetSchema',
                request_serializer=entdb__pb2.GetSchemaRequest.SerializeToString,
                response_deserializer=entdb__pb2.GetSchemaResponse.FromString,
                )
        self.ListTenants = channel.unary_unary(
                '/entdb.v1.EntDBService/ListTenants',
                request_serializer=entdb__pb2.ListTenantsRequest.SerializeToString,
                response_deserializer=entdb__pb2.ListTenantsResponse.FromString,
                )
        self.ListMailboxUsers = channel.unary_unary(
                '/entdb.v1.EntDBService/ListMailboxUsers',
                request_serializer=entdb__pb2.ListMailboxUsersRequest.SerializeToString,
                response_deserializer=entdb__pb2.ListMailboxUsersResponse.FromString,
                )


class EntDBServiceServicer(object):
    """EntDB Service - Main API for database operations
    """

    def ExecuteAtomic(self, request, context):
        """Execute an atomic transaction containing one or more operations
        Returns a receipt that can be used to check application status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetReceiptStatus(self, request, context):
        """Get the status of a previously submitted transaction
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetNode(self, request, context):
        """Get a node by ID
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetNodes(self, request, context):
        """Get multiple nodes by IDs
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def QueryNodes(self, request, context):
        """Query nodes by type with optional filters
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEdgesFrom(self, request, context):
        """Get outgoing edges from a node
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEdgesTo(self, request, context):
        """Get incoming edges to a node
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SearchMailbox(self, request, context):
        """Search mailbox with full-text search
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMailbox(self, request, context):
        """Get mailbox items for a user
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Health(self, request, context):
        """Get server health and status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSchema(self, request, context):
        """Get schema information
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListTenants(self, request, context):
        """List all tenants that have data
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListMailboxUsers(self, request, context):
        """List mailbox users for a tenant
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_EntDBServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ExecuteAtomic': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecuteAtomic,
                    request_deserializer=entdb__pb2.ExecuteAtomicRequest.FromString,
                    response_serializer=entdb__pb2.ExecuteAtomicResponse.SerializeToString,
            ),
            'GetReceiptStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetReceiptStatus,
                    request_deserializer=entdb__pb2.GetReceiptStatusRequest.FromString,
                    response_serializer=entdb__pb2.GetReceiptStatusResponse.SerializeToString,
            ),
            'GetNode': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNode,
                    request_deserializer=entdb__pb2.GetNodeRequest.FromString,
                    response_serializer=entdb__pb2.GetNodeResponse.SerializeToString,
            ),
            'GetNodes': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNodes,
                    request_deserializer=entdb__pb2.GetNodesRequest.FromString,
                    response_serializer=entdb__pb2.GetNodesResponse.SerializeToString,
            ),
            'QueryNodes': grpc.unary_unary_rpc_method_handler(
                    servicer.QueryNodes,
                    request_deserializer=entdb__pb2.QueryNodesRequest.FromString,
                    response_serializer=entdb__pb2.QueryNodesResponse.SerializeToString,
            ),
            'GetEdgesFrom': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEdgesFrom,
                    request_deserializer=entdb__pb2.GetEdgesRequest.FromString,
                    response_serializer=entdb__pb2.GetEdgesResponse.SerializeToString,
            ),
            'GetEdgesTo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEdgesTo,
                    request_deserializer=entdb__pb2.GetEdgesRequest.FromString,
                    response_serializer=entdb__pb2.GetEdgesResponse.SerializeToString,
            ),
            'SearchMailbox': grpc.unary_unary_rpc_method_handler(
                    servicer.SearchMailbox,
                    request_deserializer=entdb__pb2.SearchMailboxRequest.FromString,
                    response_serializer=entdb__pb2.SearchMailboxResponse.SerializeToString,
            ),
            'GetMailbox': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMailbox,
                    request_deserializer=entdb__pb2.GetMailboxRequest.FromString,
                    response_serializer=entdb__pb2.GetMailboxResponse.SerializeToString,
            ),
            'Health': grpc.unary_unary_rpc_method_handler(
                    servicer.Health,
                    request_deserializer=entdb__pb2.HealthRequest.FromString,
                    response_serializer=entdb__pb2.HealthResponse.SerializeToString,
            ),
            'GetSchema': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSchema,
                    request_deserializer=entdb__pb2.GetSchemaRequest.FromString,
                    response_serializer=entdb__pb2.GetSchemaResponse.SerializeToString,
            ),
            'ListTenants': grpc.unary_unary_rpc_method_handler(
                    servicer.ListTenants,
                    request_deserializer=entdb__pb2.ListTenantsRequest.FromString,
                    response_serializer=entdb__pb2.ListTenantsResponse.SerializeToString,
            ),
            'ListMailboxUsers': grpc.unary_unary_rpc_method_handler(
                    servicer.ListMailboxUsers,
                    request_deserializer=entdb__pb2.ListMailboxUsersRequest.FromString,
                    response_serializer=entdb__pb2.ListMailboxUsersResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'entdb.v1.EntDBService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class EntDBService(object):
    """EntDB Service - Main API for database operations
    """

    @staticmethod
    def ExecuteAtomic(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/ExecuteAtomic',
            entdb__pb2.ExecuteAtomicRequest.SerializeToString,
            entdb__pb2.ExecuteAtomicResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetReceiptStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/GetReceiptStatus',
            entdb__pb2.GetReceiptStatusRequest.SerializeToString,
            entdb__pb2.GetReceiptStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/GetNode',
            entdb__pb2.GetNodeRequest.SerializeToString,
            entdb__pb2.GetNodeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/GetNodes',
            entdb__pb2.GetNodesRequest.SerializeToString,
            entdb__pb2.GetNodesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def QueryNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/QueryNodes',
            entdb__pb2.QueryNodesRequest.SerializeToString,
            entdb__pb2.QueryNodesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEdgesFrom(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/GetEdgesFrom',
            entdb__pb2.GetEdgesRequest.SerializeToString,
            entdb__pb2.GetEdgesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEdgesTo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/GetEdgesTo',
            entdb__pb2.GetEdgesRequest.SerializeToString,
            entdb__pb2.GetEdgesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SearchMailbox(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/SearchMailbox',
            entdb__pb2.SearchMailboxRequest.SerializeToString,
            entdb__pb2.SearchMailboxResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetMailbox(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/GetMailbox',
            entdb__pb2.GetMailboxRequest.SerializeToString,
            entdb__pb2.GetMailboxResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Health(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/Health',
            entdb__pb2.HealthRequest.SerializeToString,
            entdb__pb2.HealthResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSchema(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/GetSchema',
            entdb__pb2.GetSchemaRequest.SerializeToString,
            entdb__pb2.GetSchemaResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListTenants(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/ListTenants',
            entdb__pb2.ListTenantsRequest.SerializeToString,
            entdb__pb2.ListTenantsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListMailboxUsers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/entdb.v1.EntDBService/ListMailboxUsers',
            entdb__pb2.ListMailboxUsersRequest.SerializeToString,
            entdb__pb2.ListMailboxUsersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""
EntDB Python SDK - Client library for EntDB database service.

This SDK provides a type-safe interface to the EntDB graph database:
- Type definitions (NodeTypeDef, EdgeTypeDef, FieldDef)
- Schema registry for type management
- DbClient for connecting to the server
- Plan builder for atomic transactions

Example:
    >>> from entdb_sdk import DbClient, NodeTypeDef, field
    >>>
    >>> # Define types
    >>> Task = NodeTypeDef(
    ...     type_id=101,
    ...     name="Task",
    ...     fields=(
    ...         field(1, "title", "str", required=True),
    ...         field(2, "status", "enum", enum_values=("todo", "done")),
    ...     ),
    ... )
    >>>
    >>> # Connect and create
    >>> async with DbClient("localhost:50051") as db:
    ...     plan = db.atomic("tenant_1", "user:42")
    ...     plan.create(Task, {"title": "My Task", "status": "todo"})
    ...     result = await plan.commit()

Invariants:
    - Type IDs are immutable after first use
    - All operations require tenant_id and actor
    - Writes are atomic per commit()

Version: see VERSION file at project root.
"""

from ._version import __version__
from .client import DbClient, Plan, Receipt
from .errors import (
    ConnectionError,
    EntDbError,
    SchemaError,
    UnknownFieldError,
    ValidationError,
)
from .registry import (
    SchemaRegistry,
    get_registry,
    register_edge_type,
    register_node_type,
)
from .schema import (
    EdgeTypeDef,
    FieldDef,
    FieldKind,
    NodeTypeDef,
    field,
)

__all__ = [
    # Version
    "__version__",
    # Schema types
    "NodeTypeDef",
    "EdgeTypeDef",
    "FieldDef",
    "FieldKind",
    "field",
    # Registry
    "SchemaRegistry",
    "get_registry",
    "register_node_type",
    "register_edge_type",
    # Client
    "DbClient",
    "Plan",
    "Receipt",
    # Errors
    "EntDbError",
    "ConnectionError",
    "ValidationError",
    "SchemaError",
    "UnknownFieldError",
]



"""
Schema types for EntDB SDK.

This module provides type definitions for the graph model:
- NodeTypeDef: Definition of a node type
- EdgeTypeDef: Definition of an edge type
- FieldDef: Individual field definition

These types are used to define the schema and enforce type safety.
IDs are stable and immutable once assigned.

Invariants:
    - type_id, edge_id, field_id never change after first release
    - Names can be renamed (ID is canonical)
    - Fields can be added but not removed
    - enum_values can only be appended

Example:
    >>> User = NodeTypeDef(
    ...     type_id=1,
    ...     name="User",
    ...     fields=(
    ...         field(1, "email", "str", required=True),
    ...         field(2, "name", "str"),
    ...     ),
    ... )
"""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field as dataclass_field
from enum import Enum
from typing import Any


class FieldKind(Enum):
    """Supported field types."""

    STRING = "str"
    INTEGER = "int"
    FLOAT = "float"
    BOOLEAN = "bool"
    TIMESTAMP = "timestamp"
    JSON = "json"
    BYTES = "bytes"
    ENUM = "enum"
    REFERENCE = "ref"
    LIST_STRING = "list_str"
    LIST_INT = "list_int"
    LIST_REF = "list_ref"

    @classmethod
    def from_str(cls, value: str) -> FieldKind:
        """Convert string to FieldKind."""
        for kind in cls:
            if kind.value == value:
                return kind
        raise ValueError(f"Invalid field kind: {value}")


@dataclass(frozen=True)
class FieldDef:
    """Field definition within a node or edge type.

    Attributes:
        field_id: Stable numeric identifier (1-65535)
        name: Human-readable name
        kind: Data type
        required: Whether field is required
        default: Default value
        enum_values: Valid values for enum type
        ref_type_id: Target type for references
        indexed: Whether to create index
        searchable: Whether to include in FTS
        deprecated: Whether field is deprecated
        description: Documentation
    """

    field_id: int
    name: str
    kind: FieldKind
    required: bool = False
    default: Any = None
    enum_values: tuple[str, ...] | None = None
    ref_type_id: int | None = None
    indexed: bool = False
    searchable: bool = False
    deprecated: bool = False
    description: str = ""

    def __post_init__(self) -> None:
        """Validate field definition."""
        if self.field_id <= 0 or self.field_id > 65535:
            raise ValueError(f"field_id must be 1-65535, got {self.field_id}")
        if not self.name:
            raise ValueError("Field name cannot be empty")
        if self.kind == FieldKind.ENUM and not self.enum_values:
            raise ValueError(f"enum_values required for ENUM field '{self.name}'")

    def validate_value(self, value: Any) -> tuple[bool, str | None]:
        """Validate a value against this field."""
        if value is None:
            if self.required:
                return False, f"Field '{self.name}' is required"
            return True, None

        if self.kind == FieldKind.ENUM:
            if self.enum_values is not None and value not in self.enum_values:
                return False, f"Field '{self.name}' must be one of {self.enum_values}"

        return True, None

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        result: dict[str, Any] = {
            "field_id": self.field_id,
            "name": self.name,
            "kind": self.kind.value,
        }
        if self.required:
            result["required"] = True
        if self.default is not None:
            result["default"] = self.default
        if self.enum_values:
            result["enum_values"] = list(self.enum_values)
        if self.ref_type_id is not None:
            result["ref_type_id"] = self.ref_type_id
        if self.indexed:
            result["indexed"] = True
        if self.searchable:
            result["searchable"] = True
        if self.deprecated:
            result["deprecated"] = True
        if self.description:
            result["description"] = self.description
        return result


def field(
    field_id: int,
    name: str,
    kind: str | FieldKind,
    *,
    required: bool = False,
    default: Any = None,
    enum_values: tuple[str, ...] | None = None,
    ref_type_id: int | None = None,
    indexed: bool = False,
    searchable: bool = False,
    deprecated: bool = False,
    description: str = "",
) -> FieldDef:
    """Convenience function to create a FieldDef.

    Args:
        field_id: Stable numeric identifier
        name: Human-readable name
        kind: Field type
        required: Whether required
        default: Default value
        enum_values: Valid enum values
        ref_type_id: Target type for references
        indexed: Create index
        searchable: Include in FTS
        deprecated: Deprecated flag
        description: Documentation

    Returns:
        FieldDef instance

    Example:
        >>> title = field(1, "title", "str", required=True)
        >>> status = field(2, "status", "enum", enum_values=("todo", "done"))
    """
    if isinstance(kind, str):
        kind = FieldKind.from_str(kind)
    return FieldDef(
        field_id=field_id,
        name=name,
        kind=kind,
        required=required,
        default=default,
        enum_values=enum_values,
        ref_type_id=ref_type_id,
        indexed=indexed,
        searchable=searchable,
        deprecated=deprecated,
        description=description,
    )


@dataclass(frozen=True)
class NodeTypeDef:
    """Definition of a node type.

    Nodes are the primary entities in the graph. Each node type
    defines the structure (fields) and behavior of nodes.

    Attributes:
        type_id: Stable numeric identifier (1-2^31)
        name: Human-readable name
        fields: Tuple of field definitions
        deprecated: Whether type is deprecated
        description: Documentation

    Example:
        >>> Task = NodeTypeDef(
        ...     type_id=101,
        ...     name="Task",
        ...     fields=(
        ...         field(1, "title", "str", required=True),
        ...         field(2, "status", "enum", enum_values=("todo", "done")),
        ...     ),
        ... )
    """

    type_id: int
    name: str
    fields: tuple[FieldDef, ...] = dataclass_field(default_factory=tuple)
    deprecated: bool = False
    description: str = ""

    def __post_init__(self) -> None:
        """Validate node type definition."""
        if self.type_id <= 0:
            raise ValueError(f"type_id must be positive, got {self.type_id}")
        if not self.name:
            raise ValueError("Node type name cannot be empty")

        # Check for duplicate field IDs
        field_ids = [f.field_id for f in self.fields]
        if len(field_ids) != len(set(field_ids)):
            raise ValueError(f"Duplicate field_id in node type '{self.name}'")

    def get_field(self, name_or_id: str | int) -> FieldDef | None:
        """Get field by name or ID."""
        for f in self.fields:
            if isinstance(name_or_id, int):
                if f.field_id == name_or_id:
                    return f
            else:
                if f.name == name_or_id:
                    return f
        return None

    def get_field_names(self) -> list[str]:
        """Get list of field names."""
        return [f.name for f in self.fields if not f.deprecated]

    def validate_payload(self, payload: dict[str, Any]) -> tuple[bool, list[str]]:
        """Validate a payload against this type."""
        errors: list[str] = []

        # Check for unknown fields
        known = {f.name for f in self.fields}
        unknown = set(payload.keys()) - known
        if unknown:
            errors.append(f"Unknown fields: {sorted(unknown)}")

        # Validate each field
        for f in self.fields:
            value = payload.get(f.name, f.default)
            is_valid, error = f.validate_value(value)
            if not is_valid and error:
                errors.append(error)

        return len(errors) == 0, errors

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        return {
            "type_id": self.type_id,
            "name": self.name,
            "fields": [f.to_dict() for f in self.fields],
            "deprecated": self.deprecated,
            "description": self.description,
        }

    def new(self, **kwargs: Any) -> dict[str, Any]:
        """Create a validated payload for this type.

        Args:
            **kwargs: Field values

        Returns:
            Validated payload dictionary

        Raises:
            TypeError: If unknown field is provided
            ValueError: If validation fails

        Example:
            >>> payload = Task.new(title="My Task", status="todo")
        """
        # Check for unknown fields
        known = {f.name for f in self.fields}
        unknown = set(kwargs.keys()) - known
        if unknown:
            suggestions = _find_suggestions(list(unknown)[0], list(known))
            msg = f"Unknown field(s): {sorted(unknown)}"
            if suggestions:
                msg += f". Did you mean: {suggestions}?"
            raise TypeError(msg)

        # Build payload with defaults
        payload: dict[str, Any] = {}
        for f in self.fields:
            if f.name in kwargs:
                payload[f.name] = kwargs[f.name]
            elif f.default is not None:
                payload[f.name] = f.default

        # Validate
        is_valid, errors = self.validate_payload(payload)
        if not is_valid:
            raise ValueError(f"Validation failed: {'; '.join(errors)}")

        return payload

    def __hash__(self) -> int:
        return hash(self.type_id)


@dataclass(frozen=True)
class EdgeTypeDef:
    """Definition of an edge type.

    Edges are unidirectional relationships between nodes.
    Each edge type defines the source/target types and properties.

    Attributes:
        edge_id: Stable numeric identifier
        name: Human-readable name
        from_type: Source node type
        to_type: Target node type
        props: Edge properties
        unique_per_from: One edge per source node
        deprecated: Whether deprecated
        description: Documentation

    Example:
        >>> AssignedTo = EdgeTypeDef(
        ...     edge_id=401,
        ...     name="AssignedTo",
        ...     from_type=Task,
        ...     to_type=User,
        ...     props=(field(1, "role", "enum", enum_values=("primary",)),),
        ... )
    """

    edge_id: int
    name: str
    from_type: NodeTypeDef | int
    to_type: NodeTypeDef | int
    props: tuple[FieldDef, ...] = dataclass_field(default_factory=tuple)
    unique_per_from: bool = False
    deprecated: bool = False
    description: str = ""

    def __post_init__(self) -> None:
        """Validate edge type definition."""
        if self.edge_id <= 0:
            raise ValueError(f"edge_id must be positive, got {self.edge_id}")
        if not self.name:
            raise ValueError("Edge type name cannot be empty")

    @property
    def from_type_id(self) -> int:
        """Get source type ID."""
        if isinstance(self.from_type, NodeTypeDef):
            return self.from_type.type_id
        return self.from_type

    @property
    def to_type_id(self) -> int:
        """Get target type ID."""
        if isinstance(self.to_type, NodeTypeDef):
            return self.to_type.type_id
        return self.to_type

    def validate_props(self, props: dict[str, Any]) -> tuple[bool, list[str]]:
        """Validate edge properties."""
        errors: list[str] = []

        known = {p.name for p in self.props}
        unknown = set(props.keys()) - known
        if unknown:
            errors.append(f"Unknown properties: {sorted(unknown)}")

        for p in self.props:
            value = props.get(p.name, p.default)
            is_valid, error = p.validate_value(value)
            if not is_valid and error:
                errors.append(error)

        return len(errors) == 0, errors

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        return {
            "edge_id": self.edge_id,
            "name": self.name,
            "from_type_id": self.from_type_id,
            "to_type_id": self.to_type_id,
            "props": [p.to_dict() for p in self.props],
            "unique_per_from": self.unique_per_from,
            "deprecated": self.deprecated,
            "description": self.description,
        }

    def __hash__(self) -> int:
        return hash(self.edge_id)


def _find_suggestions(unknown: str, known: list[str]) -> list[str]:
    """Find similar field names for suggestions."""
    suggestions = []
    unknown_lower = unknown.lower()

    for name in known:
        # Check prefix match
        if (
            (name.lower().startswith(unknown_lower[:3]) if len(unknown_lower) >= 3 else False)
            or unknown_lower in name.lower()
            or name.lower() in unknown_lower
        ):
            suggestions.append(name)

    return suggestions[:3]



"""
Error types for EntDB SDK.

This module defines all exception types raised by the SDK:
- EntDbError: Base exception
- ConnectionError: Server connection issues
- ValidationError: Payload validation failures
- SchemaError: Schema-related errors
- UnknownFieldError: Unknown field in payload

Invariants:
    - All errors inherit from EntDbError
    - Errors include context for debugging
    - Error messages are actionable
"""

from __future__ import annotations

from typing import Any


class EntDbError(Exception):
    """Base exception for all EntDB SDK errors.

    Attributes:
        message: Error message
        code: Error code for programmatic handling
        details: Additional error context
    """

    def __init__(
        self,
        message: str,
        code: str | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(message)
        self.message = message
        self.code = code or "ENTDB_ERROR"
        self.details = details or {}


class ConnectionError(EntDbError):
    """Failed to connect to EntDB server.

    Raised when:
    - Server is unreachable
    - Connection times out
    - Authentication fails
    """

    def __init__(
        self,
        message: str,
        address: str | None = None,
    ) -> None:
        super().__init__(
            message,
            code="CONNECTION_ERROR",
            details={"address": address},
        )
        self.address = address


class ValidationError(EntDbError):
    """Payload validation failed.

    Raised when:
    - Required field is missing
    - Field value has wrong type
    - Enum value is invalid
    """

    def __init__(
        self,
        message: str,
        field_name: str | None = None,
        errors: list[str] | None = None,
    ) -> None:
        super().__init__(
            message,
            code="VALIDATION_ERROR",
            details={"field": field_name, "errors": errors or []},
        )
        self.field_name = field_name
        self.errors = errors or []


class SchemaError(EntDbError):
    """Schema-related error.

    Raised when:
    - Schema fingerprint mismatch
    - Unknown type ID
    - Schema compatibility violation
    """

    def __init__(
        self,
        message: str,
        expected_fingerprint: str | None = None,
        actual_fingerprint: str | None = None,
    ) -> None:
        super().__init__(
            message,
            code="SCHEMA_ERROR",
            details={
                "expected_fingerprint": expected_fingerprint,
                "actual_fingerprint": actual_fingerprint,
            },
        )
        self.expected_fingerprint = expected_fingerprint
        self.actual_fingerprint = actual_fingerprint


class UnknownFieldError(EntDbError):
    """Unknown field in payload.

    Includes suggestions for similar field names.

    Attributes:
        field_name: The unknown field
        type_name: The type being created
        suggestions: Similar field names
    """

    def __init__(
        self,
        field_name: str,
        type_name: str,
        suggestions: list[str] | None = None,
    ) -> None:
        suggestions = suggestions or []
        msg = f"Unknown field '{field_name}' in type '{type_name}'"
        if suggestions:
            msg += f". Did you mean: {', '.join(suggestions)}?"

        super().__init__(
            msg,
            code="UNKNOWN_FIELD",
            details={
                "field_name": field_name,
                "type_name": type_name,
                "suggestions": suggestions,
            },
        )
        self.field_name = field_name
        self.type_name = type_name
        self.suggestions = suggestions


class NotFoundError(EntDbError):
    """Resource not found.

    Raised when:
    - Node doesn't exist
    - Edge doesn't exist
    - Tenant doesn't exist
    """

    def __init__(
        self,
        message: str,
        resource_type: str,
        resource_id: str,
    ) -> None:
        super().__init__(
            message,
            code="NOT_FOUND",
            details={
                "resource_type": resource_type,
                "resource_id": resource_id,
            },
        )
        self.resource_type = resource_type
        self.resource_id = resource_id


class AccessDeniedError(EntDbError):
    """Access denied.

    Raised when:
    - Actor lacks required permission
    - Node visibility doesn't include actor
    """

    def __init__(
        self,
        message: str,
        actor: str,
        resource_id: str,
        required_permission: str | None = None,
    ) -> None:
        super().__init__(
            message,
            code="ACCESS_DENIED",
            details={
                "actor": actor,
                "resource_id": resource_id,
                "required_permission": required_permission,
            },
        )
        self.actor = actor
        self.resource_id = resource_id
        self.required_permission = required_permission


class TransactionError(EntDbError):
    """Transaction failed.

    Raised when:
    - Atomic operation partially failed
    - Idempotency violation
    - Concurrent modification
    """

    def __init__(
        self,
        message: str,
        idempotency_key: str | None = None,
    ) -> None:
        super().__init__(
            message,
            code="TRANSACTION_ERROR",
            details={"idempotency_key": idempotency_key},
        )
        self.idempotency_key = idempotency_key



"""
EntDB Client for Python SDK.

This module provides the main client interface:
- DbClient: Connection to EntDB server
- Plan: Atomic transaction builder
- Receipt: Transaction receipt for status checking

Example:
    >>> async with DbClient("localhost:50051") as db:
    ...     plan = db.atomic("tenant_1", "user:42")
    ...     plan.create(Task, {"title": "My Task"})
    ...     result = await plan.commit()

Invariants:
    - All operations require tenant_id and actor
    - Writes are batched and committed atomically
    - Reads are independent of the transaction
"""

from __future__ import annotations

import json
import logging
import uuid
from dataclasses import dataclass, field
from typing import Any

from ._grpc_client import GrpcClient
from .errors import (
    ConnectionError,
    UnknownFieldError,
    ValidationError,
)
from .registry import SchemaRegistry, get_registry
from .schema import EdgeTypeDef, NodeTypeDef

logger = logging.getLogger(__name__)


@dataclass
class Receipt:
    """Transaction receipt for status tracking.

    Attributes:
        tenant_id: Tenant identifier
        idempotency_key: Unique transaction key
        stream_position: Position in WAL stream
    """

    tenant_id: str
    idempotency_key: str
    stream_position: str | None = None


@dataclass
class Node:
    """A node from the database.

    Attributes:
        tenant_id: Tenant identifier
        node_id: Unique node ID
        type_id: Node type ID
        payload: Field values
        created_at: Creation timestamp (Unix ms)
        updated_at: Last update timestamp (Unix ms)
        owner_actor: Creator
        acl: Access control list
    """

    tenant_id: str
    node_id: str
    type_id: int
    payload: dict[str, Any]
    created_at: int
    updated_at: int
    owner_actor: str
    acl: list[dict[str, str]] = field(default_factory=list)


@dataclass
class Edge:
    """An edge from the database.

    Attributes:
        tenant_id: Tenant identifier
        edge_type_id: Edge type ID
        from_node_id: Source node
        to_node_id: Target node
        props: Edge properties
        created_at: Creation timestamp
    """

    tenant_id: str
    edge_type_id: int
    from_node_id: str
    to_node_id: str
    props: dict[str, Any]
    created_at: int


@dataclass
class CommitResult:
    """Result of committing a Plan.

    Attributes:
        success: Whether commit succeeded
        receipt: Transaction receipt
        created_node_ids: IDs of created nodes
        applied: Whether event has been applied
        error: Error message if failed
    """

    success: bool
    receipt: Receipt | None = None
    created_node_ids: list[str] = field(default_factory=list)
    applied: bool = False
    error: str | None = None


class Plan:
    """Atomic transaction builder.

    A Plan collects operations to be executed atomically.
    Operations are validated locally before sending to server.

    Example:
        >>> plan = db.atomic("tenant_1", "user:42")
        >>> plan.create(Task, {"title": "My Task"}, as_="t")
        >>> plan.edge_create(AssignedTo, from_="$t.id", to={"type_id": 1, "id": "user:7"})
        >>> result = await plan.commit()
    """

    def __init__(
        self,
        client: DbClient,
        tenant_id: str,
        actor: str,
        idempotency_key: str | None = None,
    ) -> None:
        """Initialize a plan.

        Args:
            client: DbClient instance
            tenant_id: Tenant identifier
            actor: Actor performing operations
            idempotency_key: Optional unique key for deduplication
        """
        self._client = client
        self._tenant_id = tenant_id
        self._actor = actor
        self._idempotency_key = idempotency_key or str(uuid.uuid4())
        self._operations: list[dict[str, Any]] = []

    def create(
        self,
        node_type: NodeTypeDef,
        data: dict[str, Any] | None = None,
        *,
        acl: list[dict[str, str]] | None = None,
        as_: str | None = None,
        fanout_to: list[str] | None = None,
        **kwargs: Any,
    ) -> Plan:
        """Add a create_node operation.

        Args:
            node_type: Type of node to create
            data: Payload dictionary (or use kwargs)
            acl: Access control list
            as_: Alias for referencing in later operations
            fanout_to: Users to fanout to
            **kwargs: Field values (alternative to data)

        Returns:
            Self for chaining

        Raises:
            ValidationError: If payload validation fails
            UnknownFieldError: If unknown field is provided
        """
        # Merge data and kwargs
        payload = dict(data or {})
        payload.update(kwargs)

        # Validate payload
        is_valid, errors = node_type.validate_payload(payload)
        if not is_valid:
            # Check for unknown fields
            known = {f.name for f in node_type.fields}
            unknown = set(payload.keys()) - known
            if unknown:
                field_name = list(unknown)[0]
                suggestions = [n for n in known if field_name.lower() in n.lower()][:3]
                raise UnknownFieldError(field_name, node_type.name, suggestions)
            raise ValidationError("; ".join(errors), errors=errors)

        op: dict[str, Any] = {
            "create_node": {
                "type_id": node_type.type_id,
                "data_json": json.dumps(payload),
            }
        }

        if acl:
            op["create_node"]["acl_json"] = json.dumps(acl)
        if as_:
            op["create_node"]["as"] = as_
        if fanout_to:
            op["create_node"]["fanout_to"] = fanout_to

        self._operations.append(op)
        return self

    def update(
        self,
        node_type: NodeTypeDef,
        node_id: str,
        patch: dict[str, Any],
        *,
        field_mask: list[str] | None = None,
    ) -> Plan:
        """Add an update_node operation.

        Args:
            node_type: Type of node to update
            node_id: ID of node to update
            patch: Fields to update
            field_mask: Optional explicit field mask

        Returns:
            Self for chaining
        """
        op: dict[str, Any] = {
            "update_node": {
                "type_id": node_type.type_id,
                "id": node_id,
                "patch_json": json.dumps(patch),
            }
        }

        if field_mask:
            op["update_node"]["field_mask"] = field_mask

        self._operations.append(op)
        return self

    def delete(
        self,
        node_type: NodeTypeDef,
        node_id: str,
    ) -> Plan:
        """Add a delete_node operation.

        Args:
            node_type: Type of node to delete
            node_id: ID of node to delete

        Returns:
            Self for chaining
        """
        self._operations.append(
            {
                "delete_node": {
                    "type_id": node_type.type_id,
                    "id": node_id,
                }
            }
        )
        return self

    def edge_create(
        self,
        edge_type: EdgeTypeDef,
        from_: str | dict[str, Any],
        to: str | dict[str, Any],
        props: dict[str, Any] | None = None,
    ) -> Plan:
        """Add a create_edge operation.

        Args:
            edge_type: Type of edge to create
            from_: Source node (ID, alias ref, or typed ref)
            to: Target node
            props: Edge properties

        Returns:
            Self for chaining
        """
        # Validate props
        if props:
            is_valid, errors = edge_type.validate_props(props)
            if not is_valid:
                raise ValidationError("; ".join(errors), errors=errors)

        op: dict[str, Any] = {
            "create_edge": {
                "edge_id": edge_type.edge_id,
                "from": self._convert_ref(from_),
                "to": self._convert_ref(to),
            }
        }

        if props:
            op["create_edge"]["props_json"] = json.dumps(props)

        self._operations.append(op)
        return self

    def edge_delete(
        self,
        edge_type: EdgeTypeDef,
        from_: str | dict[str, Any],
        to: str | dict[str, Any],
    ) -> Plan:
        """Add a delete_edge operation.

        Args:
            edge_type: Type of edge to delete
            from_: Source node
            to: Target node

        Returns:
            Self for chaining
        """
        self._operations.append(
            {
                "delete_edge": {
                    "edge_id": edge_type.edge_id,
                    "from": self._convert_ref(from_),
                    "to": self._convert_ref(to),
                }
            }
        )
        return self

    def _convert_ref(self, ref: str | dict[str, Any]) -> dict[str, Any]:
        """Convert reference to API format."""
        if isinstance(ref, str):
            if ref.startswith("$"):
                return {"alias_ref": ref}
            return {"id": ref}
        if "type_id" in ref and "id" in ref:
            return {"typed": ref}
        return ref

    async def commit(self, wait_applied: bool = False) -> CommitResult:
        """Commit the transaction.

        Args:
            wait_applied: Whether to wait for application

        Returns:
            CommitResult indicating success/failure
        """
        if not self._operations:
            return CommitResult(success=True, created_node_ids=[])

        result = await self._client._execute(
            tenant_id=self._tenant_id,
            actor=self._actor,
            operations=self._operations,
            idempotency_key=self._idempotency_key,
            wait_applied=wait_applied,
        )

        return result


class DbClient:
    """Client for connecting to EntDB server.

    Provides a clean Python API for interacting with EntDB.
    Handles connection management and exposes high-level operations.

    Example:
        >>> async with DbClient("localhost:50051") as db:
        ...     node = await db.get(Task, "node_123", "tenant_1", "user:42")
    """

    def __init__(
        self,
        address: str,
        *,
        secure: bool = False,
        registry: SchemaRegistry | None = None,
    ) -> None:
        """Initialize client.

        Args:
            address: Server address (host:port or just host)
            secure: Whether to use TLS
            registry: Optional schema registry
        """
        # Parse address
        if ":" in address:
            host, port_str = address.rsplit(":", 1)
            port = int(port_str)
        else:
            host = address
            port = 50051  # Default gRPC port

        self._grpc = GrpcClient(host=host, port=port, secure=secure)
        self.registry = registry or get_registry()
        self._connected = False

    async def connect(self) -> None:
        """Connect to the server."""
        if self._connected:
            return

        try:
            await self._grpc.connect()
            self._connected = True
        except Exception as e:
            raise ConnectionError(
                f"Failed to connect: {e}",
                address=f"{self._grpc._host}:{self._grpc._port}",
            ) from e

    async def close(self) -> None:
        """Close the connection."""
        if self._connected:
            await self._grpc.close()
            self._connected = False

    async def __aenter__(self) -> DbClient:
        await self.connect()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()

    def atomic(
        self,
        tenant_id: str,
        actor: str,
        idempotency_key: str | None = None,
    ) -> Plan:
        """Create an atomic transaction plan.

        Args:
            tenant_id: Tenant identifier
            actor: Actor performing operations
            idempotency_key: Optional deduplication key

        Returns:
            Plan builder

        Example:
            >>> plan = db.atomic("tenant_1", "user:42")
            >>> plan.create(Task, {"title": "New Task"})
            >>> await plan.commit()
        """
        return Plan(self, tenant_id, actor, idempotency_key)

    async def get(
        self,
        node_type: NodeTypeDef,
        node_id: str,
        tenant_id: str,
        actor: str,
    ) -> Node | None:
        """Get a node by ID.

        Args:
            node_type: Expected node type
            node_id: Node identifier
            tenant_id: Tenant identifier
            actor: Actor making request

        Returns:
            Node if found, None otherwise
        """
        grpc_node = await self._grpc.get_node(
            tenant_id=tenant_id,
            actor=actor,
            type_id=node_type.type_id,
            node_id=node_id,
        )

        if grpc_node is None:
            return None

        return Node(
            tenant_id=grpc_node.tenant_id,
            node_id=grpc_node.node_id,
            type_id=grpc_node.type_id,
            payload=grpc_node.payload,
            created_at=grpc_node.created_at,
            updated_at=grpc_node.updated_at,
            owner_actor=grpc_node.owner_actor,
            acl=grpc_node.acl,
        )

    async def get_many(
        self,
        node_type: NodeTypeDef,
        node_ids: list[str],
        tenant_id: str,
        actor: str,
    ) -> tuple[list[Node], list[str]]:
        """Get multiple nodes by IDs.

        Args:
            node_type: Expected node type
            node_ids: Node identifiers
            tenant_id: Tenant identifier
            actor: Actor making request

        Returns:
            Tuple of (found nodes, missing node IDs)
        """
        grpc_nodes, missing = await self._grpc.get_nodes(
            tenant_id=tenant_id,
            actor=actor,
            type_id=node_type.type_id,
            node_ids=node_ids,
        )

        nodes = [
            Node(
                tenant_id=n.tenant_id,
                node_id=n.node_id,
                type_id=n.type_id,
                payload=n.payload,
                created_at=n.created_at,
                updated_at=n.updated_at,
                owner_actor=n.owner_actor,
                acl=n.acl,
            )
            for n in grpc_nodes
        ]

        return nodes, missing

    async def query(
        self,
        node_type: NodeTypeDef,
        tenant_id: str,
        actor: str,
        limit: int = 100,
        offset: int = 0,
    ) -> list[Node]:
        """Query nodes by type.

        Args:
            node_type: Node type to query
            tenant_id: Tenant identifier
            actor: Actor making request
            limit: Maximum nodes to return
            offset: Pagination offset

        Returns:
            List of nodes
        """
        grpc_nodes, _ = await self._grpc.query_nodes(
            tenant_id=tenant_id,
            actor=actor,
            type_id=node_type.type_id,
            limit=limit,
            offset=offset,
        )

        return [
            Node(
                tenant_id=n.tenant_id,
                node_id=n.node_id,
                type_id=n.type_id,
                payload=n.payload,
                created_at=n.created_at,
                updated_at=n.updated_at,
                owner_actor=n.owner_actor,
                acl=n.acl,
            )
            for n in grpc_nodes
        ]

    async def edges_out(
        self,
        node_id: str,
        tenant_id: str,
        actor: str,
        edge_type: EdgeTypeDef | None = None,
    ) -> list[Edge]:
        """Get outgoing edges from a node.

        Args:
            node_id: Source node ID
            tenant_id: Tenant identifier
            actor: Actor making request
            edge_type: Optional edge type filter

        Returns:
            List of edges
        """
        edge_type_id = edge_type.edge_id if edge_type else None
        grpc_edges, _ = await self._grpc.get_edges_from(
            tenant_id=tenant_id,
            actor=actor,
            node_id=node_id,
            edge_type_id=edge_type_id,
        )

        return [
            Edge(
                tenant_id=e.tenant_id,
                edge_type_id=e.edge_type_id,
                from_node_id=e.from_node_id,
                to_node_id=e.to_node_id,
                props=e.props,
                created_at=e.created_at,
            )
            for e in grpc_edges
        ]

    async def edges_in(
        self,
        node_id: str,
        tenant_id: str,
        actor: str,
        edge_type: EdgeTypeDef | None = None,
    ) -> list[Edge]:
        """Get incoming edges to a node.

        Args:
            node_id: Target node ID
            tenant_id: Tenant identifier
            actor: Actor making request
            edge_type: Optional edge type filter

        Returns:
            List of edges
        """
        edge_type_id = edge_type.edge_id if edge_type else None
        grpc_edges, _ = await self._grpc.get_edges_to(
            tenant_id=tenant_id,
            actor=actor,
            node_id=node_id,
            edge_type_id=edge_type_id,
        )

        return [
            Edge(
                tenant_id=e.tenant_id,
                edge_type_id=e.edge_type_id,
                from_node_id=e.from_node_id,
                to_node_id=e.to_node_id,
                props=e.props,
                created_at=e.created_at,
            )
            for e in grpc_edges
        ]

    async def search(
        self,
        query: str,
        user_id: str,
        tenant_id: str,
        actor: str,
        types: list[NodeTypeDef] | None = None,
        limit: int = 20,
    ) -> list[dict[str, Any]]:
        """Search user's mailbox.

        Args:
            query: Search query
            user_id: User whose mailbox to search
            tenant_id: Tenant identifier
            actor: Actor making request
            types: Optional type filter
            limit: Maximum results

        Returns:
            List of search results
        """
        source_type_ids = [t.type_id for t in types] if types else None

        return await self._grpc.search_mailbox(
            tenant_id=tenant_id,
            actor=actor,
            user_id=user_id,
            query=query,
            source_type_ids=source_type_ids,
            limit=limit,
        )

    async def health(self) -> dict[str, Any]:
        """Check server health.

        Returns:
            Health status dictionary
        """
        return await self._grpc.health()

    async def _execute(
        self,
        tenant_id: str,
        actor: str,
        operations: list[dict[str, Any]],
        idempotency_key: str,
        wait_applied: bool = False,
    ) -> CommitResult:
        """Execute atomic transaction.

        Args:
            tenant_id: Tenant identifier
            actor: Actor performing operations
            operations: List of operations
            idempotency_key: Deduplication key
            wait_applied: Wait for application

        Returns:
            CommitResult
        """
        result = await self._grpc.execute_atomic(
            tenant_id=tenant_id,
            actor=actor,
            operations=operations,
            idempotency_key=idempotency_key,
            schema_fingerprint=self.registry.fingerprint,
            wait_applied=wait_applied,
        )

        if not result.success:
            return CommitResult(
                success=False,
                error=result.error,
            )

        receipt = None
        if result.receipt:
            receipt = Receipt(
                tenant_id=result.receipt.tenant_id,
                idempotency_key=result.receipt.idempotency_key,
                stream_position=result.receipt.stream_position,
            )

        return CommitResult(
            success=True,
            receipt=receipt,
            created_node_ids=result.created_node_ids,
            applied=result.applied,
        )

    async def get_receipt_status(
        self,
        tenant_id: str,
        idempotency_key: str,
    ) -> str:
        """Get receipt status.

        Args:
            tenant_id: Tenant identifier
            idempotency_key: Transaction key

        Returns:
            Status string (PENDING, APPLIED, FAILED)
        """
        return await self._grpc.get_receipt_status(tenant_id, idempotency_key)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: entdb.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0b\x65ntdb.proto\x12\x08\x65ntdb.v1\"D\n\x0eRequestContext\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\r\n\x05\x61\x63tor\x18\x02 \x01(\t\x12\x10\n\x08trace_id\x18\x03 \x01(\t\"N\n\x07Receipt\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\x17\n\x0fidempotency_key\x18\x02 \x01(\t\x12\x17\n\x0fstream_position\x18\x03 \x01(\t\"\xce\x01\n\x14\x45xecuteAtomicRequest\x12)\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x18.entdb.v1.RequestContext\x12\x17\n\x0fidempotency_key\x18\x02 \x01(\t\x12\x1a\n\x12schema_fingerprint\x18\x03 \x01(\t\x12\'\n\noperations\x18\x04 \x03(\x0b\x32\x13.entdb.v1.Operation\x12\x14\n\x0cwait_applied\x18\x05 \x01(\x08\x12\x17\n\x0fwait_timeout_ms\x18\x06 \x01(\x05\"\xfc\x01\n\tOperation\x12-\n\x0b\x63reate_node\x18\x01 \x01(\x0b\x32\x16.entdb.v1.CreateNodeOpH\x00\x12-\n\x0bupdate_node\x18\x02 \x01(\x0b\x32\x16.entdb.v1.UpdateNodeOpH\x00\x12-\n\x0b\x64\x65lete_node\x18\x03 \x01(\x0b\x32\x16.entdb.v1.DeleteNodeOpH\x00\x12-\n\x0b\x63reate_edge\x18\x04 \x01(\x0b\x32\x16.entdb.v1.CreateEdgeOpH\x00\x12-\n\x0b\x64\x65lete_edge\x18\x05 \x01(\x0b\x32\x16.entdb.v1.DeleteEdgeOpH\x00\x42\x04\n\x02op\"o\n\x0c\x43reateNodeOp\x12\x0f\n\x07type_id\x18\x01 \x01(\x05\x12\n\n\x02id\x18\x02 \x01(\t\x12\x11\n\tdata_json\x18\x03 \x01(\t\x12\x10\n\x08\x61\x63l_json\x18\x04 \x01(\t\x12\n\n\x02\x61s\x18\x05 \x01(\t\x12\x11\n\tfanout_to\x18\x06 \x03(\t\"S\n\x0cUpdateNodeOp\x12\x0f\n\x07type_id\x18\x01 \x01(\x05\x12\n\n\x02id\x18\x02 \x01(\t\x12\x12\n\npatch_json\x18\x03 \x01(\t\x12\x12\n\nfield_mask\x18\x04 \x03(\t\"+\n\x0c\x44\x65leteNodeOp\x12\x0f\n\x07type_id\x18\x01 \x01(\x05\x12\n\n\x02id\x18\x02 \x01(\t\"s\n\x0c\x43reateEdgeOp\x12\x0f\n\x07\x65\x64ge_id\x18\x01 \x01(\x05\x12\x1f\n\x04\x66rom\x18\x02 \x01(\x0b\x32\x11.entdb.v1.NodeRef\x12\x1d\n\x02to\x18\x03 \x01(\x0b\x32\x11.entdb.v1.NodeRef\x12\x12\n\nprops_json\x18\x04 \x01(\t\"_\n\x0c\x44\x65leteEdgeOp\x12\x0f\n\x07\x65\x64ge_id\x18\x01 \x01(\x05\x12\x1f\n\x04\x66rom\x18\x02 \x01(\x0b\x32\x11.entdb.v1.NodeRef\x12\x1d\n\x02to\x18\x03 \x01(\x0b\x32\x11.entdb.v1.NodeRef\"\\\n\x07NodeRef\x12\x0c\n\x02id\x18\x01 \x01(\tH\x00\x12\x13\n\talias_ref\x18\x02 \x01(\tH\x00\x12\'\n\x05typed\x18\x03 \x01(\x0b\x32\x16.entdb.v1.TypedNodeRefH\x00\x42\x05\n\x03ref\"+\n\x0cTypedNodeRef\x12\x0f\n\x07type_id\x18\x01 \x01(\x05\x12\n\n\x02id\x18\x02 \x01(\t\"\xba\x01\n\x15\x45xecuteAtomicResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\"\n\x07receipt\x18\x02 \x01(\x0b\x32\x11.entdb.v1.Receipt\x12\r\n\x05\x65rror\x18\x03 \x01(\t\x12\x12\n\nerror_code\x18\x04 \x01(\t\x12\x18\n\x10\x63reated_node_ids\x18\x05 \x03(\t\x12/\n\x0e\x61pplied_status\x18\x06 \x01(\x0e\x32\x17.entdb.v1.ReceiptStatus\"]\n\x17GetReceiptStatusRequest\x12)\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x18.entdb.v1.RequestContext\x12\x17\n\x0fidempotency_key\x18\x02 \x01(\t\"R\n\x18GetReceiptStatusResponse\x12\'\n\x06status\x18\x01 \x01(\x0e\x32\x17.entdb.v1.ReceiptStatus\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"]\n\x0eGetNodeRequest\x12)\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x18.entdb.v1.RequestContext\x12\x0f\n\x07type_id\x18\x02 \x01(\x05\x12\x0f\n\x07node_id\x18\x03 \x01(\t\">\n\x0fGetNodeResponse\x12\x1c\n\x04node\x18\x01 \x01(\x0b\x32\x0e.entdb.v1.Node\x12\r\n\x05\x66ound\x18\x02 \x01(\x08\"_\n\x0fGetNodesRequest\x12)\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x18.entdb.v1.RequestContext\x12\x0f\n\x07type_id\x18\x02 \x01(\x05\x12\x10\n\x08node_ids\x18\x03 \x03(\t\"F\n\x10GetNodesResponse\x12\x1d\n\x05nodes\x18\x01 \x03(\x0b\x32\x0e.entdb.v1.Node\x12\x13\n\x0bmissing_ids\x18\x02 \x03(\t\"\xa9\x01\n\x11QueryNodesRequest\x12)\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x18.entdb.v1.RequestContext\x12\x0f\n\x07type_id\x18\x02 \x01(\x05\x12\x13\n\x0b\x66ilter_json\x18\x03 \x01(\t\x12\r\n\x05limit\x18\x04 \x01(\x05\x12\x0e\n\x06offset\x18\x05 \x01(\x05\x12\x10\n\x08order_by\x18\x06 \x01(\t\x12\x12\n\ndescending\x18\x07 \x01(\x08\"Z\n\x12QueryNodesResponse\x12\x1d\n\x05nodes\x18\x01 \x03(\x0b\x32\x0e.entdb.v1.Node\x12\x13\n\x0btotal_count\x18\x02 \x01(\x05\x12\x10\n\x08has_more\x18\x03 \x01(\x08\"\xa0\x01\n\x04Node\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\x0f\n\x07node_id\x18\x02 \x01(\t\x12\x0f\n\x07type_id\x18\x03 \x01(\x05\x12\x14\n\x0cpayload_json\x18\x04 \x01(\t\x12\x12\n\ncreated_at\x18\x05 \x01(\x03\x12\x12\n\nupdated_at\x18\x06 \x01(\x03\x12\x13\n\x0bowner_actor\x18\x07 \x01(\t\x12\x10\n\x08\x61\x63l_json\x18\x08 \x01(\t\"\x82\x01\n\x0fGetEdgesRequest\x12)\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x18.entdb.v1.RequestContext\x12\x0f\n\x07node_id\x18\x02 \x01(\t\x12\x14\n\x0c\x65\x64ge_type_id\x18\x03 \x01(\x05\x12\r\n\x05limit\x18\x04 \x01(\x05\x12\x0e\n\x06offset\x18\x05 \x01(\x05\"C\n\x10GetEdgesResponse\x12\x1d\n\x05\x65\x64ges\x18\x01 \x03(\x0b\x32\x0e.entdb.v1.Edge\x12\x10\n\x08has_more\x18\x02 \x01(\x08\"\x81\x01\n\x04\x45\x64ge\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\x14\n\x0c\x65\x64ge_type_id\x18\x02 \x01(\x05\x12\x14\n\x0c\x66rom_node_id\x18\x03 \x01(\t\x12\x12\n\nto_node_id\x18\x04 \x01(\t\x12\x12\n\nprops_json\x18\x05 \x01(\t\x12\x12\n\ncreated_at\x18\x06 \x01(\x03\"\x99\x01\n\x14SearchMailboxRequest\x12)\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x18.entdb.v1.RequestContext\x12\x0f\n\x07user_id\x18\x02 \x01(\t\x12\r\n\x05query\x18\x03 \x01(\t\x12\x17\n\x0fsource_type_ids\x18\x04 \x03(\x05\x12\r\n\x05limit\x18\x05 \x01(\x05\x12\x0e\n\x06offset\x18\x06 \x01(\x05\"Y\n\x15SearchMailboxResponse\x12.\n\x07results\x18\x01 \x03(\x0b\x32\x1d.entdb.v1.MailboxSearchResult\x12\x10\n\x08has_more\x18\x02 \x01(\x08\"\\\n\x13MailboxSearchResult\x12#\n\x04item\x18\x01 \x01(\x0b\x32\x15.entdb.v1.MailboxItem\x12\x0c\n\x04rank\x18\x02 \x01(\x02\x12\x12\n\nhighlights\x18\x03 \x01(\t\"\xae\x01\n\x11GetMailboxRequest\x12)\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x18.entdb.v1.RequestContext\x12\x0f\n\x07user_id\x18\x02 \x01(\t\x12\x16\n\x0esource_type_id\x18\x03 \x01(\x05\x12\x11\n\tthread_id\x18\x04 \x01(\t\x12\x13\n\x0bunread_only\x18\x05 \x01(\x08\x12\r\n\x05limit\x18\x06 \x01(\x05\x12\x0e\n\x06offset\x18\x07 \x01(\x05\"b\n\x12GetMailboxResponse\x12$\n\x05items\x18\x01 \x03(\x0b\x32\x15.entdb.v1.MailboxItem\x12\x14\n\x0cunread_count\x18\x02 \x01(\x05\x12\x10\n\x08has_more\x18\x03 \x01(\x08\"\xb9\x01\n\x0bMailboxItem\x12\x0f\n\x07item_id\x18\x01 \x01(\t\x12\x0e\n\x06ref_id\x18\x02 \x01(\t\x12\x16\n\x0esource_type_id\x18\x03 \x01(\x05\x12\x16\n\x0esource_node_id\x18\x04 \x01(\t\x12\x11\n\tthread_id\x18\x05 \x01(\t\x12\n\n\x02ts\x18\x06 \x01(\x03\x12\x12\n\nstate_json\x18\x07 \x01(\t\x12\x0f\n\x07snippet\x18\x08 \x01(\t\x12\x15\n\rmetadata_json\x18\t \x01(\t\"\x0f\n\rHealthRequest\"\xa3\x01\n\x0eHealthResponse\x12\x0f\n\x07healthy\x18\x01 \x01(\x08\x12\x0f\n\x07version\x18\x02 \x01(\t\x12<\n\ncomponents\x18\x03 \x03(\x0b\x32(.entdb.v1.HealthResponse.ComponentsEntry\x1a\x31\n\x0f\x43omponentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"6\n\x10GetSchemaRequest\x12\x0f\n\x07type_id\x18\x01 \x01(\x05\x12\x11\n\ttenant_id\x18\x02 \x01(\t\"=\n\x11GetSchemaResponse\x12\x13\n\x0bschema_json\x18\x01 \x01(\t\x12\x13\n\x0b\x66ingerprint\x18\x02 \x01(\t\"\x14\n\x12ListTenantsRequest\"<\n\x13ListTenantsResponse\x12%\n\x07tenants\x18\x01 \x03(\x0b\x32\x14.entdb.v1.TenantInfo\"\x1f\n\nTenantInfo\x12\x11\n\ttenant_id\x18\x01 \x01(\t\",\n\x17ListMailboxUsersRequest\x12\x11\n\ttenant_id\x18\x01 \x01(\t\",\n\x18ListMailboxUsersResponse\x12\x10\n\x08user_ids\x18\x01 \x03(\t*~\n\rReceiptStatus\x12\x1a\n\x16RECEIPT_STATUS_UNKNOWN\x10\x00\x12\x1a\n\x16RECEIPT_STATUS_PENDING\x10\x01\x12\x1a\n\x16RECEIPT_STATUS_APPLIED\x10\x02\x12\x19\n\x15RECEIPT_STATUS_FAILED\x10\x03\x32\xd8\x07\n\x0c\x45ntDBService\x12P\n\rExecuteAtomic\x12\x1e.entdb.v1.ExecuteAtomicRequest\x1a\x1f.entdb.v1.ExecuteAtomicResponse\x12Y\n\x10GetReceiptStatus\x12!.entdb.v1.GetReceiptStatusRequest\x1a\".entdb.v1.GetReceiptStatusResponse\x12>\n\x07GetNode\x12\x18.entdb.v1.GetNodeRequest\x1a\x19.entdb.v1.GetNodeResponse\x12\x41\n\x08GetNodes\x12\x19.entdb.v1.GetNodesRequest\x1a\x1a.entdb.v1.GetNodesResponse\x12G\n\nQueryNodes\x12\x1b.entdb.v1.QueryNodesRequest\x1a\x1c.entdb.v1.QueryNodesResponse\x12\x45\n\x0cGetEdgesFrom\x12\x19.entdb.v1.GetEdgesRequest\x1a\x1a.entdb.v1.GetEdgesResponse\x12\x43\n\nGetEdgesTo\x12\x19.entdb.v1.GetEdgesRequest\x1a\x1a.entdb.v1.GetEdgesResponse\x12P\n\rSearchMailbox\x12\x1e.entdb.v1.SearchMailboxRequest\x1a\x1f.entdb.v1.SearchMailboxResponse\x12G\n\nGetMailbox\x12\x1b.entdb.v1.GetMailboxRequest\x1a\x1c.entdb.v1.GetMailboxResponse\x12;\n\x06Health\x12\x17.entdb.v1.HealthRequest\x1a\x18.entdb.v1.HealthResponse\x12\x44\n\tGetSchema\x12\x1a.entdb.v1.GetSchemaRequest\x1a\x1b.entdb.v1.GetSchemaResponse\x12J\n\x0bListTenants\x12\x1c.entdb.v1.ListTenantsRequest\x1a\x1d.entdb.v1.ListTenantsResponse\x12Y\n\x10ListMailboxUsers\x12!.entdb.v1.ListMailboxUsersRequest\x1a\".entdb.v1.ListMailboxUsersResponseB#Z!github.com/example/entdb/proto/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'entdb_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z!github.com/example/entdb/proto/v1'
  _HEALTHRESPONSE_COMPONENTSENTRY._options = None
  _HEALTHRESPONSE_COMPONENTSENTRY._serialized_options = b'8\001'
  _globals['_RECEIPTSTATUS']._serialized_start=4009
  _globals['_RECEIPTSTATUS']._serialized_end=4135
  _globals['_REQUESTCONTEXT']._serialized_start=25
  _globals['_REQUESTCONTEXT']._serialized_end=93
  _globals['_RECEIPT']._serialized_start=95
  _globals['_RECEIPT']._serialized_end=173
  _globals['_EXECUTEATOMICREQUEST']._serialized_start=176
  _globals['_EXECUTEATOMICREQUEST']._serialized_end=382
  _globals['_OPERATION']._serialized_start=385
  _globals['_OPERATION']._serialized_end=637
  _globals['_CREATENODEOP']._serialized_start=639
  _globals['_CREATENODEOP']._serialized_end=750
  _globals['_UPDATENODEOP']._serialized_start=752
  _globals['_UPDATENODEOP']._serialized_end=835
  _globals['_DELETENODEOP']._serialized_start=837
  _globals['_DELETENODEOP']._serialized_end=880
  _globals['_CREATEEDGEOP']._serialized_start=882
  _globals['_CREATEEDGEOP']._serialized_end=997
  _globals['_DELETEEDGEOP']._serialized_start=999
  _globals['_DELETEEDGEOP']._serialized_end=1094
  _globals['_NODEREF']._serialized_start=1096
  _globals['_NODEREF']._serialized_end=1188
  _globals['_TYPEDNODEREF']._serialized_start=1190
  _globals['_TYPEDNODEREF']._serialized_end=1233
  _globals['_EXECUTEATOMICRESPONSE']._serialized_start=1236
  _globals['_EXECUTEATOMICRESPONSE']._serialized_end=1422
  _globals['_GETRECEIPTSTATUSREQUEST']._serialized_start=1424
  _globals['_GETRECEIPTSTATUSREQUEST']._serialized_end=1517
  _globals['_GETRECEIPTSTATUSRESPONSE']._serialized_start=1519
  _globals['_GETRECEIPTSTATUSRESPONSE']._serialized_end=1601
  _globals['_GETNODEREQUEST']._serialized_start=1603
  _globals['_GETNODEREQUEST']._serialized_end=1696
  _globals['_GETNODERESPONSE']._serialized_start=1698
  _globals['_GETNODERESPONSE']._serialized_end=1760
  _globals['_GETNODESREQUEST']._serialized_start=1762
  _globals['_GETNODESREQUEST']._serialized_end=1857
  _globals['_GETNODESRESPONSE']._serialized_start=1859
  _globals['_GETNODESRESPONSE']._serialized_end=1929
  _globals['_QUERYNODESREQUEST']._serialized_start=1932
  _globals['_QUERYNODESREQUEST']._serialized_end=2101
  _globals['_QUERYNODESRESPONSE']._serialized_start=2103
  _globals['_QUERYNODESRESPONSE']._serialized_end=2193
  _globals['_NODE']._serialized_start=2196
  _globals['_NODE']._serialized_end=2356
  _globals['_GETEDGESREQUEST']._serialized_start=2359
  _globals['_GETEDGESREQUEST']._serialized_end=2489
  _globals['_GETEDGESRESPONSE']._serialized_start=2491
  _globals['_GETEDGESRESPONSE']._serialized_end=2558
  _globals['_EDGE']._serialized_start=2561
  _globals['_EDGE']._serialized_end=2690
  _globals['_SEARCHMAILBOXREQUEST']._serialized_start=2693
  _globals['_SEARCHMAILBOXREQUEST']._serialized_end=2846
  _globals['_SEARCHMAILBOXRESPONSE']._serialized_start=2848
  _globals['_SEARCHMAILBOXRESPONSE']._serialized_end=2937
  _globals['_MAILBOXSEARCHRESULT']._serialized_start=2939
  _globals['_MAILBOXSEARCHRESULT']._serialized_end=3031
  _globals['_GETMAILBOXREQUEST']._serialized_start=3034
  _globals['_GETMAILBOXREQUEST']._serialized_end=3208
  _globals['_GETMAILBOXRESPONSE']._serialized_start=3210
  _globals['_GETMAILBOXRESPONSE']._serialized_end=3308
  _globals['_MAILBOXITEM']._serialized_start=3311
  _globals['_MAILBOXITEM']._serialized_end=3496
  _globals['_HEALTHREQUEST']._serialized_start=3498
  _globals['_HEALTHREQUEST']._serialized_end=3513
  _globals['_HEALTHRESPONSE']._serialized_start=3516
  _globals['_HEALTHRESPONSE']._serialized_end=3679
  _globals['_HEALTHRESPONSE_COMPONENTSENTRY']._serialized_start=3630
  _globals['_HEALTHRESPONSE_COMPONENTSENTRY']._serialized_end=3679
  _globals['_GETSCHEMAREQUEST']._serialized_start=3681
  _globals['_GETSCHEMAREQUEST']._serialized_end=3735
  _globals['_GETSCHEMARESPONSE']._serialized_start=3737
  _globals['_GETSCHEMARESPONSE']._serialized_end=3798
  _globals['_LISTTENANTSREQUEST']._serialized_start=3800
  _globals['_LISTTENANTSREQUEST']._serialized_end=3820
  _globals['_LISTTENANTSRESPONSE']._serialized_start=3822
  _globals['_LISTTENANTSRESPONSE']._serialized_end=3882
  _globals['_TENANTINFO']._serialized_start=3884
  _globals['_TENANTINFO']._serialized_end=3915
  _globals['_LISTMAILBOXUSERSREQUEST']._serialized_start=3917
  _globals['_LISTMAILBOXUSERSREQUEST']._serialized_end=3961
  _globals['_LISTMAILBOXUSERSRESPONSE']._serialized_start=3963
  _globals['_LISTMAILBOXUSERSRESPONSE']._serialized_end=4007
  _globals['_ENTDBSERVICE']._serialized_start=4138
  _globals['_ENTDBSERVICE']._serialized_end=5122
# @@protoc_insertion_point(module_scope)



"""
Internal gRPC client for EntDB SDK.

This module provides the low-level gRPC communication layer.
It is internal to the SDK and should not be used directly by users.

Users should use DbClient instead, which provides a clean Python API.
"""

from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from typing import Any

import grpc
from grpc import aio as grpc_aio

from ._generated import (
    CreateEdgeOp,
    CreateNodeOp,
    DeleteEdgeOp,
    DeleteNodeOp,
    EntDBServiceStub,
    ExecuteAtomicRequest,
    GetEdgesRequest,
    GetMailboxRequest,
    GetNodeRequest,
    GetNodesRequest,
    GetReceiptStatusRequest,
    GetSchemaRequest,
    HealthRequest,
    NodeRef,
    Operation,
    QueryNodesRequest,
    # Enums
    ReceiptStatus,
    # Request types
    RequestContext,
    SearchMailboxRequest,
    TypedNodeRef,
    UpdateNodeOp,
)

logger = logging.getLogger(__name__)


@dataclass
class GrpcNode:
    """Internal node representation from gRPC response."""

    tenant_id: str
    node_id: str
    type_id: int
    payload: dict[str, Any]
    created_at: int
    updated_at: int
    owner_actor: str
    acl: list[dict[str, str]]


@dataclass
class GrpcEdge:
    """Internal edge representation from gRPC response."""

    tenant_id: str
    edge_type_id: int
    from_node_id: str
    to_node_id: str
    props: dict[str, Any]
    created_at: int


@dataclass
class GrpcReceipt:
    """Internal receipt representation from gRPC response."""

    tenant_id: str
    idempotency_key: str
    stream_position: str | None


@dataclass
class GrpcCommitResult:
    """Internal commit result from gRPC response."""

    success: bool
    receipt: GrpcReceipt | None
    created_node_ids: list[str]
    applied: bool
    error: str | None


class GrpcClient:
    """Internal gRPC client for EntDB.

    This class handles all gRPC communication with the server.
    It manages connection lifecycle and provides async methods
    for all RPC operations.

    This is an internal class - users should use DbClient instead.
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 50051,
        *,
        secure: bool = False,
        credentials: grpc.ChannelCredentials | None = None,
    ) -> None:
        """Initialize the gRPC client.

        Args:
            host: Server hostname
            port: Server port
            secure: Whether to use TLS
            credentials: Optional TLS credentials
        """
        self._host = host
        self._port = port
        self._secure = secure
        self._credentials = credentials
        self._channel: grpc_aio.Channel | None = None
        self._stub: EntDBServiceStub | None = None

    async def connect(self) -> None:
        """Establish connection to the server."""
        if self._channel is not None:
            return

        address = f"{self._host}:{self._port}"

        if self._secure:
            if self._credentials:
                self._channel = grpc_aio.secure_channel(address, self._credentials)
            else:
                self._channel = grpc_aio.secure_channel(
                    address,
                    grpc.ssl_channel_credentials(),
                )
        else:
            self._channel = grpc_aio.insecure_channel(
                address,
                options=[
                    ("grpc.max_send_message_length", 50 * 1024 * 1024),
                    ("grpc.max_receive_message_length", 50 * 1024 * 1024),
                ],
            )

        self._stub = EntDBServiceStub(self._channel)
        logger.debug(f"Connected to EntDB server at {address}")

    async def close(self) -> None:
        """Close the connection."""
        if self._channel:
            await self._channel.close()
            self._channel = None
            self._stub = None
            logger.debug("Disconnected from EntDB server")

    async def __aenter__(self) -> GrpcClient:
        await self.connect()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()

    def _ensure_connected(self) -> EntDBServiceStub:
        """Ensure we're connected and return the stub."""
        if self._stub is None:
            raise RuntimeError("Not connected. Call connect() first.")
        return self._stub

    def _make_context(self, tenant_id: str, actor: str) -> RequestContext:
        """Create a request context."""
        return RequestContext(tenant_id=tenant_id, actor=actor)

    async def execute_atomic(
        self,
        tenant_id: str,
        actor: str,
        operations: list[dict[str, Any]],
        *,
        idempotency_key: str | None = None,
        schema_fingerprint: str | None = None,
        wait_applied: bool = False,
        wait_timeout_ms: int = 30000,
    ) -> GrpcCommitResult:
        """Execute an atomic transaction.

        Args:
            tenant_id: Tenant identifier
            actor: Actor performing the operation
            operations: List of operations in SDK format
            idempotency_key: Optional idempotency key
            schema_fingerprint: Schema fingerprint for validation
            wait_applied: Whether to wait for application
            wait_timeout_ms: Timeout for wait_applied

        Returns:
            GrpcCommitResult with success/failure info
        """
        stub = self._ensure_connected()

        # Convert operations to protobuf format
        proto_ops = self._convert_operations(operations)

        request = ExecuteAtomicRequest(
            context=self._make_context(tenant_id, actor),
            idempotency_key=idempotency_key or "",
            schema_fingerprint=schema_fingerprint or "",
            operations=proto_ops,
            wait_applied=wait_applied,
            wait_timeout_ms=wait_timeout_ms,
        )

        try:
            response = await stub.ExecuteAtomic(request)

            receipt = None
            if response.receipt and response.receipt.idempotency_key:
                receipt = GrpcReceipt(
                    tenant_id=response.receipt.tenant_id,
                    idempotency_key=response.receipt.idempotency_key,
                    stream_position=response.receipt.stream_position or None,
                )

            return GrpcCommitResult(
                success=response.success,
                receipt=receipt,
                created_node_ids=list(response.created_node_ids),
                applied=response.applied_status == ReceiptStatus.RECEIPT_STATUS_APPLIED,
                error=response.error if response.error else None,
            )
        except grpc.RpcError as e:
            return GrpcCommitResult(
                success=False,
                receipt=None,
                created_node_ids=[],
                applied=False,
                error=str(e),
            )

    def _convert_operations(self, operations: list[dict[str, Any]]) -> list[Operation]:
        """Convert SDK operations to protobuf format."""
        result = []

        for op in operations:
            proto_op = Operation()

            if "create_node" in op:
                create = op["create_node"]
                create_op = CreateNodeOp(
                    type_id=create.get("type_id", 0),
                    id=create.get("id", ""),
                    data_json=create.get("data_json", "{}"),
                    acl_json=create.get("acl_json", ""),
                )
                if create.get("as"):
                    # protobuf field is named 'as' but Python keyword conflict
                    setattr(create_op, "as", create["as"])
                if create.get("fanout_to"):
                    create_op.fanout_to.extend(create["fanout_to"])
                proto_op.create_node.CopyFrom(create_op)

            elif "update_node" in op:
                update = op["update_node"]
                update_op = UpdateNodeOp(
                    type_id=update.get("type_id", 0),
                    id=update.get("id", ""),
                    patch_json=update.get("patch_json", "{}"),
                )
                if update.get("field_mask"):
                    update_op.field_mask.extend(update["field_mask"])
                proto_op.update_node.CopyFrom(update_op)

            elif "delete_node" in op:
                delete = op["delete_node"]
                proto_op.delete_node.CopyFrom(
                    DeleteNodeOp(
                        type_id=delete.get("type_id", 0),
                        id=delete.get("id", ""),
                    )
                )

            elif "create_edge" in op:
                create = op["create_edge"]
                create_op = CreateEdgeOp(
                    edge_id=create.get("edge_id", 0),
                    props_json=create.get("props_json", "{}"),
                )
                getattr(create_op, "from").CopyFrom(self._convert_node_ref(create.get("from", {})))
                create_op.to.CopyFrom(self._convert_node_ref(create.get("to", {})))
                proto_op.create_edge.CopyFrom(create_op)

            elif "delete_edge" in op:
                delete = op["delete_edge"]
                delete_op = DeleteEdgeOp(edge_id=delete.get("edge_id", 0))
                getattr(delete_op, "from").CopyFrom(self._convert_node_ref(delete.get("from", {})))
                delete_op.to.CopyFrom(self._convert_node_ref(delete.get("to", {})))
                proto_op.delete_edge.CopyFrom(delete_op)

            result.append(proto_op)

        return result

    def _convert_node_ref(self, ref: dict[str, Any]) -> NodeRef:
        """Convert node reference to protobuf format."""
        node_ref = NodeRef()

        if "id" in ref:
            node_ref.id = ref["id"]
        elif "alias_ref" in ref:
            node_ref.alias_ref = ref["alias_ref"]
        elif "typed" in ref:
            node_ref.typed.CopyFrom(
                TypedNodeRef(
                    type_id=ref["typed"].get("type_id", 0),
                    id=ref["typed"].get("id", ""),
                )
            )

        return node_ref

    async def get_receipt_status(
        self,
        tenant_id: str,
        idempotency_key: str,
    ) -> str:
        """Get the status of a transaction receipt.

        Returns:
            Status string: "PENDING", "APPLIED", "FAILED", or "UNKNOWN"
        """
        stub = self._ensure_connected()

        request = GetReceiptStatusRequest(
            context=self._make_context(tenant_id, "system"),
            idempotency_key=idempotency_key,
        )

        response = await stub.GetReceiptStatus(request)

        status_map = {
            ReceiptStatus.RECEIPT_STATUS_PENDING: "PENDING",
            ReceiptStatus.RECEIPT_STATUS_APPLIED: "APPLIED",
            ReceiptStatus.RECEIPT_STATUS_FAILED: "FAILED",
            ReceiptStatus.RECEIPT_STATUS_UNKNOWN: "UNKNOWN",
        }
        return status_map.get(response.status, "UNKNOWN")

    async def get_node(
        self,
        tenant_id: str,
        actor: str,
        type_id: int,
        node_id: str,
    ) -> GrpcNode | None:
        """Get a node by ID."""
        stub = self._ensure_connected()

        request = GetNodeRequest(
            context=self._make_context(tenant_id, actor),
            type_id=type_id,
            node_id=node_id,
        )

        response = await stub.GetNode(request)

        if not response.found:
            return None

        node = response.node
        return GrpcNode(
            tenant_id=node.tenant_id,
            node_id=node.node_id,
            type_id=node.type_id,
            payload=json.loads(node.payload_json) if node.payload_json else {},
            created_at=node.created_at,
            updated_at=node.updated_at,
            owner_actor=node.owner_actor,
            acl=json.loads(node.acl_json) if node.acl_json else [],
        )

    async def get_nodes(
        self,
        tenant_id: str,
        actor: str,
        type_id: int,
        node_ids: list[str],
    ) -> tuple[list[GrpcNode], list[str]]:
        """Get multiple nodes by IDs.

        Returns:
            Tuple of (found nodes, missing node IDs)
        """
        stub = self._ensure_connected()

        request = GetNodesRequest(
            context=self._make_context(tenant_id, actor),
            type_id=type_id,
            node_ids=node_ids,
        )

        response = await stub.GetNodes(request)

        nodes = [
            GrpcNode(
                tenant_id=n.tenant_id,
                node_id=n.node_id,
                type_id=n.type_id,
                payload=json.loads(n.payload_json) if n.payload_json else {},
                created_at=n.created_at,
                updated_at=n.updated_at,
                owner_actor=n.owner_actor,
                acl=json.loads(n.acl_json) if n.acl_json else [],
            )
            for n in response.nodes
        ]

        return nodes, list(response.missing_ids)

    async def query_nodes(
        self,
        tenant_id: str,
        actor: str,
        type_id: int,
        *,
        limit: int = 100,
        offset: int = 0,
        filter_json: str | None = None,
        order_by: str | None = None,
        descending: bool = False,
    ) -> tuple[list[GrpcNode], bool]:
        """Query nodes by type.

        Returns:
            Tuple of (nodes, has_more)
        """
        stub = self._ensure_connected()

        request = QueryNodesRequest(
            context=self._make_context(tenant_id, actor),
            type_id=type_id,
            limit=limit,
            offset=offset,
            filter_json=filter_json or "",
            order_by=order_by or "",
            descending=descending,
        )

        response = await stub.QueryNodes(request)

        nodes = [
            GrpcNode(
                tenant_id=n.tenant_id,
                node_id=n.node_id,
                type_id=n.type_id,
                payload=json.loads(n.payload_json) if n.payload_json else {},
                created_at=n.created_at,
                updated_at=n.updated_at,
                owner_actor=n.owner_actor,
                acl=json.loads(n.acl_json) if n.acl_json else [],
            )
            for n in response.nodes
        ]

        return nodes, response.has_more

    async def get_edges_from(
        self,
        tenant_id: str,
        actor: str,
        node_id: str,
        edge_type_id: int | None = None,
        limit: int = 100,
    ) -> tuple[list[GrpcEdge], bool]:
        """Get outgoing edges from a node.

        Returns:
            Tuple of (edges, has_more)
        """
        stub = self._ensure_connected()

        request = GetEdgesRequest(
            context=self._make_context(tenant_id, actor),
            node_id=node_id,
            edge_type_id=edge_type_id or 0,
            limit=limit,
        )

        response = await stub.GetEdgesFrom(request)

        edges = [
            GrpcEdge(
                tenant_id=e.tenant_id,
                edge_type_id=e.edge_type_id,
                from_node_id=e.from_node_id,
                to_node_id=e.to_node_id,
                props=json.loads(e.props_json) if e.props_json else {},
                created_at=e.created_at,
            )
            for e in response.edges
        ]

        return edges, response.has_more

    async def get_edges_to(
        self,
        tenant_id: str,
        actor: str,
        node_id: str,
        edge_type_id: int | None = None,
        limit: int = 100,
    ) -> tuple[list[GrpcEdge], bool]:
        """Get incoming edges to a node.

        Returns:
            Tuple of (edges, has_more)
        """
        stub = self._ensure_connected()

        request = GetEdgesRequest(
            context=self._make_context(tenant_id, actor),
            node_id=node_id,
            edge_type_id=edge_type_id or 0,
            limit=limit,
        )

        response = await stub.GetEdgesTo(request)

        edges = [
            GrpcEdge(
                tenant_id=e.tenant_id,
                edge_type_id=e.edge_type_id,
                from_node_id=e.from_node_id,
                to_node_id=e.to_node_id,
                props=json.loads(e.props_json) if e.props_json else {},
                created_at=e.created_at,
            )
            for e in response.edges
        ]

        return edges, response.has_more

    async def search_mailbox(
        self,
        tenant_id: str,
        actor: str,
        user_id: str,
        query: str,
        *,
        source_type_ids: list[int] | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> list[dict[str, Any]]:
        """Search user's mailbox with full-text search."""
        stub = self._ensure_connected()

        request = SearchMailboxRequest(
            context=self._make_context(tenant_id, actor),
            user_id=user_id,
            query=query,
            limit=limit,
            offset=offset,
        )
        if source_type_ids:
            request.source_type_ids.extend(source_type_ids)

        response = await stub.SearchMailbox(request)

        return [
            {
                "item": {
                    "item_id": r.item.item_id,
                    "ref_id": r.item.ref_id,
                    "source_type_id": r.item.source_type_id,
                    "source_node_id": r.item.source_node_id,
                    "thread_id": r.item.thread_id,
                    "ts": r.item.ts,
                    "state": json.loads(r.item.state_json) if r.item.state_json else {},
                    "snippet": r.item.snippet,
                    "metadata": json.loads(r.item.metadata_json) if r.item.metadata_json else {},
                },
                "rank": r.rank,
                "highlights": r.highlights,
            }
            for r in response.results
        ]

    async def get_mailbox(
        self,
        tenant_id: str,
        actor: str,
        user_id: str,
        *,
        source_type_id: int | None = None,
        thread_id: str | None = None,
        unread_only: bool = False,
        limit: int = 50,
        offset: int = 0,
    ) -> tuple[list[dict[str, Any]], int, bool]:
        """Get mailbox items for a user.

        Returns:
            Tuple of (items, unread_count, has_more)
        """
        stub = self._ensure_connected()

        request = GetMailboxRequest(
            context=self._make_context(tenant_id, actor),
            user_id=user_id,
            source_type_id=source_type_id or 0,
            thread_id=thread_id or "",
            unread_only=unread_only,
            limit=limit,
            offset=offset,
        )

        response = await stub.GetMailbox(request)

        items = [
            {
                "item_id": item.item_id,
                "ref_id": item.ref_id,
                "source_type_id": item.source_type_id,
                "source_node_id": item.source_node_id,
                "thread_id": item.thread_id,
                "ts": item.ts,
                "state": json.loads(item.state_json) if item.state_json else {},
                "snippet": item.snippet,
                "metadata": json.loads(item.metadata_json) if item.metadata_json else {},
            }
            for item in response.items
        ]

        return items, response.unread_count, response.has_more

    async def health(self) -> dict[str, Any]:
        """Get server health status."""
        stub = self._ensure_connected()

        response = await stub.Health(HealthRequest())

        return {
            "healthy": response.healthy,
            "version": response.version,
            "components": dict(response.components),
        }

    async def get_schema(self, type_id: int | None = None) -> dict[str, Any]:
        """Get schema information."""
        stub = self._ensure_connected()

        request = GetSchemaRequest(type_id=type_id or 0)
        response = await stub.GetSchema(request)

        return {
            "schema": json.loads(response.schema_json) if response.schema_json else {},
            "fingerprint": response.fingerprint,
        }



"""
Schema registry for EntDB SDK.

This module provides a local schema registry for:
- Registering node and edge types
- Type lookup by ID or name
- Schema fingerprinting

The registry is frozen at startup to prevent runtime modifications.

Example:
    >>> from entdb_sdk import get_registry, NodeTypeDef, field
    >>>
    >>> User = NodeTypeDef(type_id=1, name="User", fields=(field(1, "email", "str"),))
    >>> registry = get_registry()
    >>> registry.register_node_type(User)
"""

from __future__ import annotations

import hashlib
import json
import threading
from collections.abc import Iterator

from .schema import EdgeTypeDef, NodeTypeDef

# Global registry
_global_registry: SchemaRegistry | None = None
_registry_lock = threading.Lock()


class RegistryFrozenError(Exception):
    """Registry is frozen and cannot be modified."""

    pass


class DuplicateRegistrationError(Exception):
    """Type or edge with this ID is already registered."""

    pass


class SchemaRegistry:
    """Local schema registry for type management.

    The registry stores all node and edge type definitions and provides
    lookup by ID or name. It can be frozen to prevent modifications.

    Example:
        >>> registry = SchemaRegistry()
        >>> registry.register_node_type(User)
        >>> registry.register_edge_type(AssignedTo)
        >>> registry.freeze()
    """

    def __init__(self) -> None:
        """Initialize empty registry."""
        self._node_types: dict[int, NodeTypeDef] = {}
        self._edge_types: dict[int, EdgeTypeDef] = {}
        self._node_types_by_name: dict[str, NodeTypeDef] = {}
        self._edge_types_by_name: dict[str, EdgeTypeDef] = {}
        self._frozen = False
        self._fingerprint: str | None = None
        self._lock = threading.Lock()

    @property
    def frozen(self) -> bool:
        """Whether registry is frozen."""
        return self._frozen

    @property
    def fingerprint(self) -> str | None:
        """Schema fingerprint (available after freeze)."""
        return self._fingerprint

    def register_node_type(self, node_type: NodeTypeDef) -> None:
        """Register a node type.

        Args:
            node_type: NodeTypeDef to register

        Raises:
            RegistryFrozenError: If registry is frozen
            DuplicateRegistrationError: If type_id already registered
        """
        with self._lock:
            if self._frozen:
                raise RegistryFrozenError("Cannot register: registry is frozen")

            if node_type.type_id in self._node_types:
                existing = self._node_types[node_type.type_id]
                raise DuplicateRegistrationError(
                    f"type_id {node_type.type_id} already registered as '{existing.name}'"
                )

            self._node_types[node_type.type_id] = node_type
            self._node_types_by_name[node_type.name] = node_type

    def register_edge_type(self, edge_type: EdgeTypeDef) -> None:
        """Register an edge type.

        Args:
            edge_type: EdgeTypeDef to register

        Raises:
            RegistryFrozenError: If registry is frozen
            DuplicateRegistrationError: If edge_id already registered
        """
        with self._lock:
            if self._frozen:
                raise RegistryFrozenError("Cannot register: registry is frozen")

            if edge_type.edge_id in self._edge_types:
                existing = self._edge_types[edge_type.edge_id]
                raise DuplicateRegistrationError(
                    f"edge_id {edge_type.edge_id} already registered as '{existing.name}'"
                )

            self._edge_types[edge_type.edge_id] = edge_type
            self._edge_types_by_name[edge_type.name] = edge_type

    def get_node_type(self, type_id_or_name: int | str) -> NodeTypeDef | None:
        """Get node type by ID or name."""
        if isinstance(type_id_or_name, int):
            return self._node_types.get(type_id_or_name)
        return self._node_types_by_name.get(type_id_or_name)

    def get_edge_type(self, edge_id_or_name: int | str) -> EdgeTypeDef | None:
        """Get edge type by ID or name."""
        if isinstance(edge_id_or_name, int):
            return self._edge_types.get(edge_id_or_name)
        return self._edge_types_by_name.get(edge_id_or_name)

    def node_types(self) -> Iterator[NodeTypeDef]:
        """Iterate over all node types."""
        yield from self._node_types.values()

    def edge_types(self) -> Iterator[EdgeTypeDef]:
        """Iterate over all edge types."""
        yield from self._edge_types.values()

    def freeze(self) -> str:
        """Freeze registry and compute fingerprint.

        Returns:
            Schema fingerprint

        Raises:
            RegistryFrozenError: If already frozen
        """
        with self._lock:
            if self._frozen:
                raise RegistryFrozenError("Registry is already frozen")

            self._fingerprint = self._compute_fingerprint()
            self._frozen = True
            return self._fingerprint

    def _compute_fingerprint(self) -> str:
        """Compute SHA-256 fingerprint."""
        schema_dict = self.to_dict()
        canonical = json.dumps(schema_dict, sort_keys=True, separators=(",", ":"))
        hash_bytes = hashlib.sha256(canonical.encode("utf-8")).hexdigest()
        return f"sha256:{hash_bytes}"

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        return {
            "node_types": [
                self._node_types[tid].to_dict() for tid in sorted(self._node_types.keys())
            ],
            "edge_types": [
                self._edge_types[eid].to_dict() for eid in sorted(self._edge_types.keys())
            ],
        }

    def to_json(self, indent: int | None = 2) -> str:
        """Convert to JSON string."""
        return json.dumps(self.to_dict(), indent=indent, sort_keys=True)


def get_registry() -> SchemaRegistry:
    """Get the global schema registry."""
    global _global_registry
    with _registry_lock:
        if _global_registry is None:
            _global_registry = SchemaRegistry()
        return _global_registry


def register_node_type(node_type: NodeTypeDef) -> None:
    """Register a node type in the global registry."""
    get_registry().register_node_type(node_type)


def register_edge_type(edge_type: EdgeTypeDef) -> None:
    """Register an edge type in the global registry."""
    get_registry().register_edge_type(edge_type)


def reset_registry() -> None:
    """Reset the global registry (for testing only)."""
    global _global_registry
    with _registry_lock:
        _global_registry = None



"""
Payload validation for EntDB SDK.

This module provides validation utilities:
- Field-level validation
- Payload validation against types
- Helpful error messages with suggestions

Invariants:
    - Validation errors are deterministic
    - Error messages include context for fixing
    - Unknown fields suggest similar valid fields
"""

from __future__ import annotations

from difflib import get_close_matches
from typing import Any

from .errors import UnknownFieldError, ValidationError
from .schema import FieldKind, NodeTypeDef


def validate_payload(
    node_type: NodeTypeDef,
    payload: dict[str, Any],
) -> tuple[bool, list[str]]:
    """Validate payload against node type.

    Args:
        node_type: Node type to validate against
        payload: Payload to validate

    Returns:
        Tuple of (is_valid, list_of_errors)
    """
    errors: list[str] = []

    # Check for unknown fields
    known_fields = {f.name for f in node_type.fields}
    unknown = set(payload.keys()) - known_fields
    if unknown:
        for field_name in unknown:
            suggestions = get_close_matches(field_name, list(known_fields), n=3)
            if suggestions:
                errors.append(f"Unknown field '{field_name}'. Did you mean: {suggestions}?")
            else:
                errors.append(f"Unknown field '{field_name}'")

    # Validate each field
    for field_def in node_type.fields:
        value = payload.get(field_def.name, field_def.default)

        # Check required
        if value is None and field_def.required:
            errors.append(f"Field '{field_def.name}' is required")
            continue

        if value is None:
            continue

        # Type-specific validation
        error = _validate_field_value(field_def.name, field_def.kind, value, field_def.enum_values)
        if error:
            errors.append(error)

    return len(errors) == 0, errors


def _validate_field_value(
    name: str,
    kind: FieldKind,
    value: Any,
    enum_values: tuple[str, ...] | None = None,
) -> str | None:
    """Validate a single field value.

    Returns error message if invalid, None if valid.
    """
    if kind == FieldKind.STRING:
        if not isinstance(value, str):
            return f"Field '{name}' must be a string, got {type(value).__name__}"

    elif kind == FieldKind.INTEGER:
        if not isinstance(value, int) or isinstance(value, bool):
            return f"Field '{name}' must be an integer, got {type(value).__name__}"

    elif kind == FieldKind.FLOAT:
        if not isinstance(value, (int, float)) or isinstance(value, bool):
            return f"Field '{name}' must be a number, got {type(value).__name__}"

    elif kind == FieldKind.BOOLEAN:
        if not isinstance(value, bool):
            return f"Field '{name}' must be a boolean, got {type(value).__name__}"

    elif kind == FieldKind.TIMESTAMP:
        if not isinstance(value, int) or value < 0:
            return f"Field '{name}' must be a positive integer timestamp"

    elif kind == FieldKind.ENUM:
        if not isinstance(value, str):
            return f"Field '{name}' must be a string, got {type(value).__name__}"
        if enum_values and value not in enum_values:
            return f"Field '{name}' must be one of {enum_values}, got '{value}'"

    elif kind == FieldKind.LIST_STRING:
        if not isinstance(value, list):
            return f"Field '{name}' must be a list, got {type(value).__name__}"
        for i, item in enumerate(value):
            if not isinstance(item, str):
                return f"Field '{name}[{i}]' must be a string"

    elif kind == FieldKind.LIST_INT:
        if not isinstance(value, list):
            return f"Field '{name}' must be a list, got {type(value).__name__}"
        for i, item in enumerate(value):
            if not isinstance(item, int) or isinstance(item, bool):
                return f"Field '{name}[{i}]' must be an integer"

    elif kind == FieldKind.REFERENCE:
        if not isinstance(value, dict):
            return f"Field '{name}' must be a reference object"
        if "type_id" not in value or "id" not in value:
            return f"Field '{name}' must have 'type_id' and 'id'"

    return None


def validate_or_raise(
    node_type: NodeTypeDef,
    payload: dict[str, Any],
) -> None:
    """Validate payload and raise if invalid.

    Args:
        node_type: Node type to validate against
        payload: Payload to validate

    Raises:
        UnknownFieldError: If unknown field is provided
        ValidationError: If validation fails
    """
    # Check unknown fields first (for better error messages)
    known_fields = {f.name for f in node_type.fields}
    unknown = set(payload.keys()) - known_fields

    if unknown:
        field_name = list(unknown)[0]
        suggestions = get_close_matches(field_name, list(known_fields), n=3)
        raise UnknownFieldError(field_name, node_type.name, suggestions)

    is_valid, errors = validate_payload(node_type, payload)
    if not is_valid:
        raise ValidationError(
            f"Validation failed for {node_type.name}: {'; '.join(errors)}",
            errors=errors,
        )


def suggest_fields(
    partial: str,
    node_type: NodeTypeDef,
    limit: int = 5,
) -> list[str]:
    """Suggest field names based on partial input.

    Args:
        partial: Partial field name
        node_type: Node type to suggest from
        limit: Maximum suggestions

    Returns:
        List of suggested field names
    """
    known = [f.name for f in node_type.fields if not f.deprecated]
    matches = get_close_matches(partial, known, n=limit)

    # Also include prefix matches
    prefix_matches = [n for n in known if n.lower().startswith(partial.lower())]

    # Combine and deduplicate
    all_matches = list(dict.fromkeys(matches + prefix_matches))
    return all_matches[:limit]
